"""Tests for cross-section shapes."""

import math

import numpy as np
import pytest
from sectionproperties.analysis import Section

from blueprints.structural_sections.cross_section_annular_sector import AnnularSectorCrossSection


class TestAnnularSectorCrossSection:
    """Tests for the AnnularSectorCrossSection class."""

    def test_area(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the area property of the AnnularSectorCrossSection class."""
        expected_area = (annular_sector_cross_section.outer_radius**2 - annular_sector_cross_section.inner_radius**2) * np.pi / 4
        assert annular_sector_cross_section.area == pytest.approx(expected=expected_area, rel=1e-6)

    def test_width(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the width property of the AnnularSectorCrossSection class."""
        expected_width = 110.0
        assert annular_sector_cross_section.width == pytest.approx(expected=expected_width, rel=1e-6)

    def test_height(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the height property of the AnnularSectorCrossSection class."""
        expected_height = 110.0
        assert annular_sector_cross_section.height == pytest.approx(expected=expected_height, rel=1e-6)

    def test_perimeter(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the perimeter property of the AnnularSectorCrossSection class."""
        angle_radians = 90.0 * (np.pi / 180)
        expected_perimeter = angle_radians * (110.0 + 90.0) + 2 * 20.0
        assert annular_sector_cross_section.perimeter == pytest.approx(expected=expected_perimeter, rel=1e-6)

    def test_centroid(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the centroid property of the AnnularSectorCrossSection class."""
        centroid = annular_sector_cross_section.centroid
        halve_angle_radians = np.radians(annular_sector_cross_section.end_angle - annular_sector_cross_section.start_angle) / 2
        centroid_radius = (2 * np.sin(halve_angle_radians) / (3 * halve_angle_radians)) * (
            (annular_sector_cross_section.outer_radius**3 - annular_sector_cross_section.inner_radius**3)
            / (annular_sector_cross_section.outer_radius**2 - annular_sector_cross_section.inner_radius**2)
        )
        centroid_angle = np.radians((annular_sector_cross_section.start_angle + annular_sector_cross_section.end_angle) / 2)
        centroid_x = annular_sector_cross_section.x + centroid_radius * np.cos(np.radians(90) - centroid_angle)
        centroid_y = annular_sector_cross_section.y + centroid_radius * np.sin(np.radians(90) - centroid_angle)
        assert centroid.x == pytest.approx(expected=centroid_x, rel=1e-6)
        assert centroid.y == pytest.approx(expected=centroid_y, rel=1e-6)

    def test_moments_of_inertia(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the moments of inertia properties of the AnnularSectorCrossSection class."""
        assert annular_sector_cross_section.moment_of_inertia_about_y == annular_sector_cross_section.moment_of_inertia_about_z

    def test_section_moduli(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the section moduli properties of the AnnularSectorCrossSection class."""
        assert annular_sector_cross_section.elastic_section_modulus_about_y_positive == pytest.approx(
            expected=annular_sector_cross_section.elastic_section_modulus_about_z_positive, rel=1e-6
        )
        assert annular_sector_cross_section.elastic_section_modulus_about_y_negative == pytest.approx(
            expected=annular_sector_cross_section.elastic_section_modulus_about_z_negative, rel=1e-6
        )

    def test_plastic_section_moduli(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the plastic section moduli properties of the AnnularSectorCrossSection class."""
        assert annular_sector_cross_section.plastic_section_modulus_about_y is None
        assert annular_sector_cross_section.plastic_section_modulus_about_z is None

    def test_polygon(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the polygon property of the AnnularSectorCrossSection class."""
        assert annular_sector_cross_section.polygon.is_valid
        assert len(annular_sector_cross_section.polygon.exterior.coords) > 0

    def test_section(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the section object of the AnnularSectorCrossSection class."""
        section = annular_sector_cross_section.section()
        assert isinstance(section, Section)

    def test_geometry(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the geometry property of the AnnularSectorCrossSection class."""
        geometry = annular_sector_cross_section.geometry()
        assert geometry is not None

    def test_invalid_radius(self) -> None:
        """Test initialization with an invalid radius value."""
        with pytest.raises(ValueError, match="Radius must be zero or positive"):
            AnnularSectorCrossSection(
                name="InvalidRadius",
                inner_radius=-10.0,
                thickness=20.0,
                start_angle=0.0,
                end_angle=90.0,
                x=0.0,
                y=0.0,
            )

    def test_invalid_thickness(self) -> None:
        """Test initialization with an invalid thickness value."""
        with pytest.raises(ValueError, match="Thickness must be a positive value"):
            AnnularSectorCrossSection(
                name="InvalidThickness",
                inner_radius=100.0,
                thickness=-20.0,
                start_angle=0.0,
                end_angle=90.0,
                x=0.0,
                y=0.0,
            )

    def test_invalid_end_angle_smaller_than_start_angle(self) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="InvalidEndAngle",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=90.0,
                end_angle=0.0,
                x=0.0,
                y=0.0,
            )

    @pytest.mark.parametrize(
        ("start_angle", "end_angle"),
        [
            (0, 360),
            (0, 720),
            (-50, 360),
            (0, 400),
        ],
    )
    def test_total_angle_greater_or_equal_than_360(self, start_angle: float, end_angle: float) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="Invalid total angle, greater than 360 degrees",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=start_angle,
                end_angle=end_angle,
                x=0.0,
                y=0.0,
            )

    def test_invalid_start_angle(self) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="InvalidEndAngle",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=-683,
                end_angle=270,
                x=0.0,
                y=0.0,
            )

    @pytest.mark.parametrize(
        ("start_angle", "end_angle", "expected_area"),
        [
            (0, 90, 3141.5926),
            (0, 180, 3141.5926 * 2),
            (0, 270, 3141.5926 * 3),
            (0, 359.9999999, 3141.5926 * 4),
            (90, 180, 3141.5926),
            (90, 270, 3141.5926 * 2),
            (90, 360, 3141.5926 * 3),
            (-90, 269.999999, 3141.5926 * 4),
            (-180, 90, 3141.5926 * 3),
            (-360, -270, 3141.5926),
        ],
    )
    def test_area_at_different_angles(self, start_angle: float, end_angle: float, expected_area: float) -> None:
        """Test the area property of the AnnularSectorCrossSection class at different angles.
        A full donut (0 to 360 degrees) should have an area of 0.
        a quarter donut (0 to 90 degrees) should have an area of pi * (outer_radius^2 - inner_radius^2) / 4.
        """
        annular_sector = AnnularSectorCrossSection(
            inner_radius=90.0,
            thickness=20.0,
            start_angle=start_angle,
            end_angle=end_angle,
            x=100.0,
            y=250.0,
        )
        assert annular_sector.area == pytest.approx(expected=expected_area, rel=1e-6)

    def test_moments_of_inertia_approximation_with_rectangle(self) -> None:
        """Test the moments of inertia by approximating the annular sector with a rectangle."""
        annular_sector = AnnularSectorCrossSection(
            inner_radius=1000,  # mm
            thickness=0.1,  # mm
            start_angle=-2,  # degrees
            end_angle=2,  # degrees
            x=0,  # mm
            y=0,  # mm
        )
        # Approximate the moments of inertia using a rectangle
        angle_radians = math.radians(annular_sector.end_angle - annular_sector.start_angle)
        approximate_width = angle_radians * annular_sector.radius_centerline
        approximate_height = annular_sector.thickness

        # Moment of inertia about the z-axis
        approximate_moi_z = (approximate_height * approximate_width**3) / 12
        assert annular_sector.moment_of_inertia_about_z == pytest.approx(expected=approximate_moi_z, rel=1e-3)

        annular_sector = AnnularSectorCrossSection(
            inner_radius=1000,  # mm
            thickness=0.1,  # mm
            start_angle=88,  # degrees
            end_angle=92,  # degrees
            x=0,  # mm
            y=0,  # mm
        )
        # Approximate the moments of inertia using a rectangle
        angle_radians = math.radians(annular_sector.end_angle - annular_sector.start_angle)
        approximate_width = annular_sector.thickness
        approximate_height = angle_radians * annular_sector.radius_centerline

        # Moment of inertia about the z-axis
        approximate_moi_y = (approximate_height**3 * approximate_width) / 12
        assert annular_sector.moment_of_inertia_about_y == pytest.approx(expected=approximate_moi_y, rel=1e-3)

    def test_section_properties(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the section properties of the HexagonalCrossSection class."""
        section_properties = annular_sector_cross_section.section_properties()
        assert section_properties.area == pytest.approx(expected=annular_sector_cross_section.area, rel=1e-2)
        assert section_properties.perimeter == pytest.approx(expected=annular_sector_cross_section.perimeter, rel=1e-2)
        assert section_properties.cx == pytest.approx(expected=annular_sector_cross_section.centroid.x, rel=1e-2)
        assert section_properties.cy == pytest.approx(expected=annular_sector_cross_section.centroid.y, rel=1e-2)
        assert section_properties.ixx_c == pytest.approx(expected=annular_sector_cross_section.moment_of_inertia_about_y, rel=1e-2)
        assert section_properties.iyy_c == pytest.approx(expected=annular_sector_cross_section.moment_of_inertia_about_z, rel=1e-2)
        assert section_properties.zxx_plus == pytest.approx(expected=annular_sector_cross_section.elastic_section_modulus_about_y_positive, rel=1e-2)
        assert section_properties.zyy_plus == pytest.approx(expected=annular_sector_cross_section.elastic_section_modulus_about_z_positive, rel=1e-2)
        assert section_properties.zxx_minus == pytest.approx(expected=annular_sector_cross_section.elastic_section_modulus_about_y_negative, rel=1e-2)
        assert section_properties.zyy_minus == pytest.approx(expected=annular_sector_cross_section.elastic_section_modulus_about_z_negative, rel=1e-2)

    def test_section_properties_359_degrees(self, annular_sector_cross_section_359_degrees: AnnularSectorCrossSection) -> None:
        """Test the section properties of the HexagonalCrossSection class with 359 degrees."""
        section_properties = annular_sector_cross_section_359_degrees.section_properties()
        assert section_properties.area == pytest.approx(expected=annular_sector_cross_section_359_degrees.area, rel=1e-2)
        assert section_properties.perimeter == pytest.approx(expected=annular_sector_cross_section_359_degrees.perimeter, rel=1e-2)
        assert section_properties.cx == pytest.approx(expected=annular_sector_cross_section_359_degrees.centroid.x, rel=1e-2)
        assert section_properties.cy == pytest.approx(expected=annular_sector_cross_section_359_degrees.centroid.y, rel=1e-2)
        assert section_properties.ixx_c == pytest.approx(expected=annular_sector_cross_section_359_degrees.moment_of_inertia_about_y, rel=1e-2)
        assert section_properties.iyy_c == pytest.approx(expected=annular_sector_cross_section_359_degrees.moment_of_inertia_about_z, rel=1e-2)
        assert section_properties.zxx_plus == pytest.approx(
            expected=annular_sector_cross_section_359_degrees.elastic_section_modulus_about_y_positive, rel=1e-2
        )
        assert section_properties.zyy_plus == pytest.approx(
            expected=annular_sector_cross_section_359_degrees.elastic_section_modulus_about_z_positive, rel=1e-2
        )
        assert section_properties.zxx_minus == pytest.approx(
            expected=annular_sector_cross_section_359_degrees.elastic_section_modulus_about_y_negative, rel=1e-2
        )
        assert section_properties.zyy_minus == pytest.approx(
            expected=annular_sector_cross_section_359_degrees.elastic_section_modulus_about_z_negative, rel=1e-2
        )



"""Formula 3.29 from EN 1992-1-1:2004: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, HOURS, PERCENTAGE


class Form3Dot29RatioLossOfPreStressClass2(Formula):
    """Class representing formula 3.29 for the calculation of the ratio between loss of pre-stress and initial pre-stress of class 2."""

    label = "3.29"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        rho_1000: PERCENTAGE,
        mu: DIMENSIONLESS,
        t: HOURS,
    ) -> None:
        r"""[$\frac{\Delta \sigma_{pr}}{\sigma_{pi}}$] Ratio between loss of pre-stress and initial pre-stress for class 2. [$-$].

        EN 1992-1-1:2004 art.3.3.2(7) - Formula (3.29)

        Parameters
        ----------
        rho_1000 : PERCENTAGE
            [$\rho_{1000}$] Value of relaxation loss at 1000h after prestressing at an average temperature of 20 degrees Celsius [$%$]
        mu : DIMENSIONLESS
            [$\mu$] Ratio between initial pre-stress and characteristic tensile strength [$-$]
            = [$\sigma_{pi} / f_{pk}$]
            Use your own implementation of this formula or use sub_formula_3_28_39_30 class SubForm3Dot282930Mu.
        t : HOURS
            [$t$] Time after prestressing [$hours$]

        Returns
        -------
        None
        """
        super().__init__()
        self.rho_1000 = rho_1000
        self.mu = mu
        self.t = t

    @staticmethod
    def _evaluate(
        rho_1000: PERCENTAGE,
        mu: DIMENSIONLESS,
        t: HOURS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if rho_1000 < 0:
            raise ValueError(f"Invalid rho_1000: {rho_1000}. rho_1000 cannot be negative")
        if t < 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative")
        return 0.66 * rho_1000 * np.exp(9.1 * mu) * (t / 1000) ** (0.75 * (1 - mu)) * 10**-5

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.29."""
        return LatexFormula(
            return_symbol=r"\frac{\Delta \sigma_{pr}}{\sigma_{pl}}",
            result=f"{self:.6f}",
            equation=r"0.66 \cdot \rho_{1000} \cdot e^{9.1 \cdot \mu} \left( \frac{t}{1000} \right)^{0.75 \cdot (1 - \mu)} \cdot 10^{-5}",
            numeric_equation=(
                rf"0.66 \cdot {self.rho_1000:.3f} \cdot e^{{9.1 \cdot {self.mu:.3f}}} \left( \frac{{{self.t:.3f}}}{{1000}} \right)"
                rf"^{{0.75 \cdot (1 - {self.mu:.3f})}} \cdot 10^{{-5}}"
            ),
            comparison_operator_label="=",
        )



"""Generate the quick reference page using the README.md file and fixing the cross links."""

import re
from pathlib import Path

import mkdocs_gen_files

root = Path(__file__).parents[2]
readme_file = root / "README.md"


def replace_links(readme_content: str) -> str:
    """Replace the relative links to the overview page by removing the .md extension."""
    readme_content = readme_content.replace("docs/objects_overview/", "")
    # replace the relative links to the overview page by removing the .md extension
    readme_content = re.sub(r"/\w+\.md", lambda match: match.group(0)[:-3], readme_content)
    # replace blueprints/code references by links to the API reference
    readme_content = readme_content.replace('"blueprints/codes', '"../API reference/codes')
    # add 'https://github.com/Blueprints-org/blueprints' prefix to file in [link](file) if file is not a URL
    return re.sub(r"\[([^\]]+)\]\((?!http)([^)]+)\)", r"[\1](https://github.com/Blueprints-org/blueprints/blob/main/\2)", readme_content)


with open(readme_file, encoding="utf-8") as fd:
    ## extract the  ## Quick Reference to Blueprint's Objects section
    readme_content = fd.read()
    pattern = re.compile(r"## Quick Reference to Blueprint's Objects.*?(?=## )", re.DOTALL)

    match = pattern.search(readme_content)

    if match:
        quick_reference_content = match.group(0).replace("##", "#")
        quick_reference_content = replace_links(quick_reference_content)
    else:
        raise ValueError("No match found for reference table")

with mkdocs_gen_files.open("objects_overview/index.md", "w") as fd:
    # hide navigation in the index page
    fd.write(quick_reference_content)



"""Tests for cross-section shapes."""

import math

import pytest
from sectionproperties.analysis import Section
from shapely import Polygon

from blueprints.structural_sections.cross_section_circle import CircularCrossSection


class TestCircularCrossSection:
    """Tests for the CircularCrossSection class."""

    def test_polygon(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the polygon property of the CircularCrossSection class."""
        assert isinstance(circular_cross_section.polygon, Polygon)

    def test_area(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the area property of the CircularCrossSection class."""
        assert circular_cross_section.area == pytest.approx(expected=31415.92653, rel=1e-6)

    def test_perimeter(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the perimeter property of the CircularCrossSection class."""
        assert circular_cross_section.perimeter == pytest.approx(expected=628.31853, rel=1e-6)

    def test_centroid(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the centroid property of the CircularCrossSection class."""
        centroid = circular_cross_section.centroid
        assert (centroid.x, centroid.y) == pytest.approx(expected=(100.0, 250.0), rel=1e-6)

    def test_wrong_input(self) -> None:
        """Test the wrong input for the CircularCrossSection class."""
        with pytest.raises(ValueError):
            CircularCrossSection(name="Circle", diameter=-200.0, x=0.0, y=0.0)

    def test_moments_of_inertia(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the moments of inertia properties of the CircularCrossSection class."""
        expected_moi = (1 / 64) * math.pi * 200**4
        assert circular_cross_section.moment_of_inertia_about_y == pytest.approx(expected=expected_moi, rel=1e-6)
        assert circular_cross_section.moment_of_inertia_about_z == pytest.approx(expected=expected_moi, rel=1e-6)

    def test_section_moduli(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the section moduli properties of the CircularCrossSection class."""
        expected_section_modulus = (1 / 32) * math.pi * 200**3
        assert circular_cross_section.elastic_section_modulus_about_y_positive == pytest.approx(expected=expected_section_modulus, rel=1e-6)
        assert circular_cross_section.elastic_section_modulus_about_z_positive == pytest.approx(expected=expected_section_modulus, rel=1e-6)
        assert circular_cross_section.elastic_section_modulus_about_y_negative == pytest.approx(expected=expected_section_modulus, rel=1e-6)
        assert circular_cross_section.elastic_section_modulus_about_z_negative == pytest.approx(expected=expected_section_modulus, rel=1e-6)

    def test_plastic_section_moduli(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the plastic section moduli properties of the CircularCrossSection class."""
        expected_plastic_modulus = (1 / 6) * 200**3
        assert circular_cross_section.plastic_section_modulus_about_y == pytest.approx(expected=expected_plastic_modulus, rel=1e-6)
        assert circular_cross_section.plastic_section_modulus_about_z == pytest.approx(expected=expected_plastic_modulus, rel=1e-6)

    def test_radius(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the radius property of the CircularCrossSection class."""
        assert circular_cross_section.radius == pytest.approx(expected=100.0, rel=1e-6)

    def test_geometry_bounds(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the bounds of the geometry property."""
        bounds = circular_cross_section.polygon.bounds
        assert bounds == pytest.approx(expected=(0.0, 150.0, 200.0, 350.0), rel=1e-6)

    def test_section(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the section object of the CircularCrossSection class."""
        section = circular_cross_section.section()
        assert isinstance(section, Section)

    def test_geometry(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the geometry property of the CircularCrossSection class."""
        geometry = circular_cross_section.geometry()
        assert geometry is not None

    def test_section_properties(self, circular_cross_section: CircularCrossSection) -> None:
        """Test the section properties of the CircularCrossSection class."""
        section_properties = circular_cross_section.section_properties()
        assert section_properties.area == pytest.approx(expected=circular_cross_section.area, rel=1e-2)
        assert section_properties.perimeter == pytest.approx(expected=circular_cross_section.perimeter, rel=1e-2)
        assert section_properties.cx == pytest.approx(expected=circular_cross_section.centroid.x, rel=1e-2)
        assert section_properties.cy == pytest.approx(expected=circular_cross_section.centroid.y, rel=1e-2)
        assert section_properties.ixx_c == pytest.approx(expected=circular_cross_section.moment_of_inertia_about_y, rel=1e-2)
        assert section_properties.iyy_c == pytest.approx(expected=circular_cross_section.moment_of_inertia_about_z, rel=1e-2)
        assert section_properties.zxx_plus == pytest.approx(expected=circular_cross_section.elastic_section_modulus_about_y_positive, rel=1e-2)
        assert section_properties.zyy_plus == pytest.approx(expected=circular_cross_section.elastic_section_modulus_about_z_positive, rel=1e-2)
        assert section_properties.zxx_minus == pytest.approx(expected=circular_cross_section.elastic_section_modulus_about_y_negative, rel=1e-2)
        assert section_properties.zyy_minus == pytest.approx(expected=circular_cross_section.elastic_section_modulus_about_z_negative, rel=1e-2)
        assert section_properties.sxx == pytest.approx(expected=circular_cross_section.plastic_section_modulus_about_y, rel=1e-2)
        assert section_properties.syy == pytest.approx(expected=circular_cross_section.plastic_section_modulus_about_z, rel=1e-2)
